#!/usr/bin/perl -w
# mysar.pl - see header for version
# Record mysql status and variables in a DB table
# 
# Updated by Pythian, Inc. 2009
# www.pythian.com
# 
# Copyright (C) 2009 Gerardo Narvaja
#
# This utility is based ideas from mysqltuner and statpack
# This code is intended to ultimately be included on a new utility
# being developed @ Pythian or for other users to buid other
# utilities on top of it.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This project would not be possible without help from:
#	Sheeri Cabral
#	Tim Procter
#
# Project page: http://launchpad.net/mysar

use strict;
use warnings;
use diagnostics;
use Getopt::Long;
use DBI;

# dummy variable to eliminate code while debugging
my $bigassif = 1;

# Set up a few variables for use in the script
my $mysarversion = "0.5-alpha";

# Set defaults
my %opt = (
		"host" => 0,
		"socket" => 0,
		"port" => 0,
		"user" => 0,
		"pass" => 0,
		"variables" => 0,
		"status" => 0,
		"varfile" => 0,
		"statfile" => 0,
		"database" => 0,
		"purge" => -1,		# '0' is a valid value
	);
	
# Gather the options from the command line
GetOptions(\%opt,
		'host=s',
		'socket=s',
		'port=i',
		'user=s',
		'pass=s',
		'database=s',
		'variables',
		'status',
		'varfile=s',
		'statfile=s',
		'help',
		'purge'
	);

if (defined $opt{'help'} && $opt{'help'} == 1) { usage(); }

sub usage {
	# Shown when --help option is passed
	print "\n".
		"   mysar $mysarversion - MySQL Variables gathering utility modeled after 'sar'\n".
		"   Maintained by Gerardo Narvaja - Licensed under GPL v2\n".
		"\n".
		"   Options List:\n".
		"      --host <hostname>    Connect to a remote host to perform tests (default: localhost)\n".
		"      --socket <socket>    Use a different socket for a local connection\n".
		"      --port <port>        Port to use for connection (default: 3306)\n".
		"      --user <username>    (required) Username to use for authentication\n".
		"      --pass <password>    (required) Password to use for authentication\n".
		"      --variables          Gather the global variables values\n".
		"      --status             Gather the global status values\n".
		"                           You need to specify at least one of the options above\n".
		"      --varfile            File with the list of variables to monitor. (default: all)\n".
		"      --statfile           File with the list of status variables to monitor (default: all)\n".
		"                           For variables/status specification you may use '%' as wildcard\n".
		"      --database           Specifies which schema contains the tables 'value_var' & 'value_stat'\n". 
		"                           (default: mysar)\n". 
		"      --purge              Specify the number of days to keep in the table(s) being updated.\n\n".
		"   Use 'perldoc mysar.pl' for more detailed information\n".
		"\n";
	exit;
}

# DB variables
my $dsn = "DBI:mysql";			# connection string seed
my $dbh;				# DB  handler
my $sth;				# DB stat
my $my_database = "mysar";		# Which database schema to use
my $my_purge = 30;			# How many days to keep

sub mysql_setup {
	my $my_user = "";
	my $my_pass = "";
	my $my_host = "localhost";
	my $my_socket = 0;
	my $my_port = "3306";

	# Pick the connection parameters
	# user and password are mandatory
	if ($opt{user} ne 0) {
		$my_user = "$opt{user}";
		print "User: $my_user\n";
	} else {
		print "ERROR: You need to specify user name w/ --user\n";
		usage();
		exit 1;
	}
	if ($opt{pass} ne 0) {
		$my_pass = "$opt{pass}";
	} else { 
		print "ERROR: You need to specify user password w/ --pass\n";
		usage();
		exit 1;
	}
	
	# Optional connection parameters
	if ($opt{database} ne 0) {
		$my_database = "$opt{database}";
		print "Using tables $my_database.value_var & $my_database.value_stat\n";
	}
	if ($opt{host} ne 0) {
		$my_host = "$opt{host}";
		print "The values are going to be retrieved from: $my_host\n";
	}
	if ($opt{port} ne 0) {
		$my_port = "$opt{port}";
		print "Using port $my_port\n";
	}
	# Use default or specified values
	$dsn = "$dsn:$my_database;host=$my_host;port=$my_port";

	# Test for socket specification
	if ($opt{socket} ne 0) {
		$my_socket = "$opt{socket}";
		$dsn = "$dsn;mysql_socket=$my_socket";
	}
	
	# Connect to the DB
	$dbh = DBI->connect($dsn, $my_user, $my_pass)
		or die "ERROR: Unable to connect to MySQL server $my_host w/ user: $my_user\n".
                       "   --> $!\n";

	# Set the purge days
	if ($opt{purge} ne -1) {
		$my_purge = "$opt{purge}";
		print "Purging values older than $my_purge days";
	}
}

# Function to check if the tables exist
sub tbl_exists($) {
	my (@tblexists);
	my $table = $_[0];

	eval {
		$dbh->{PrintError} = 0;
		if($dbh->{RaiseError} == 0){
			$dbh->{RaiseError} = 1;
			print "Failed to find table $my_database.$table\n";
        		}

		};
	my $sth = $dbh->prepare("SHOW TABLES LIKE '$table'") or return 0;
	$sth->execute() or return 0;
	@tblexists = $sth->fetchrow_array() or return 0;
	if ($tblexists[0]) { return 1;} else {return 0}
	$sth->finish();

	if ($@) {
		print "returning 0 for $table\n";
        	return 0;
	}
	print "returning 1 for $table\n";
	return 1;
}

# Verify the tables and create them if they don't exist
sub verify_table($) {
	my $table = $_[0];

	if (tbl_exists($table) eq 1) { return; }

	print "*** Creating $my_database.$table \n";
	if ($table eq "value_var") {
		$dbh->do("CREATE TABLE IF NOT EXISTS `value_var` ( ".
			"`value_var_id` int(11) NOT NULL auto_increment, ".
			"`value_var_timestamp` timestamp NOT NULL default CURRENT_TIMESTAMP, ".
			"`value_var_name` varchar(64) NOT NULL, ".
			"`value_var_number` bigint(20) default NULL, ".
			"`value_var_string` varchar(256) default NULL, ".
			"PRIMARY KEY  (`value_var_id`) ".
			") ENGINE=InnoDB")
				or die "ERROR: Failed to create $table\n";
	}
	if ($table eq "value_stat") {
		$dbh->do("CREATE TABLE IF NOT EXISTS `value_stat` ( ".
			"`value_stat_id` int(11) NOT NULL auto_increment, ".
			"`value_stat_timestamp` timestamp NOT NULL default CURRENT_TIMESTAMP, ".
			"`value_stat_name` varchar(64) NOT NULL, ".
			"`value_stat_number` bigint(20) default NULL, ".
			"`value_stat_delta` bigint(20) default NULL," .
			"`value_stat_string` varchar(256) default NULL, ".
			"PRIMARY KEY  (`value_stat_id`) ".
			") ENGINE=InnoDB ")
				or die "ERROR: Failed to create $table\n";
	}
}


# Flags to determine what to pick from the database
my $my_variables = 0;
my $my_all_variables = 0;
my $my_status = 0;
my $my_all_status = 0;

sub get_status_vars {
	# Retrieve the list of (global) status
	my $ts_par = $_[0];
	my (@mystatlist_values, @delta_value);
	$delta_value[0] = 0;		# eliminate a warning
	my ($stat_name, $stat_value, $in_stat);

	print "*** Retrieving status variables from the DB\n";

	# if sepecified purge old values
	if ($my_purge ne 0) {
		print "***> purge before $my_purge days\n";
		$sth = $dbh->prepare("DELETE FROM value_stat ".
			"WHERE DATE(value_stat_timestamp) = DATE(SUBDATE(NOW(), INTERVAL ? DAY))")
				or die "ERROR: DELETE FROM value_stat on prepare\n";
		$sth->execute($my_purge)
			or die "ERROR: DELETE FROM value_stat on execute\n";
	} # if $my_purge

	if ($opt{'statfile'} ne 0) {
		print "Using status list '$opt{statfile}'\n";
		open($in_stat,"< $opt{statfile}") or die "$opt{statfile}: $!\n";
	} else {
		print "Will retrieve all status variables\n";
		$my_all_status = 1;
	} # if $opt()


	if ($my_all_status eq 1) {
		$sth = $dbh->prepare("SHOW /* !50001 GLOBAL */ STATUS")
			or die "ERROR: SHOW (GLOBAL) STATUS failed\n";
		$sth->execute()
			or die "ERROR: SHOW (GLOBAL) STATUS failed on execute\n";
		while ( @mystatlist_values = $sth->fetchrow_array() ) {
			$stat_name = $mystatlist_values[0];
			$stat_value = $mystatlist_values[1];
			
			if  (($stat_value =~ /\D/) || (length($stat_value) eq 0)) {
				$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
					" value_stat_string) VALUES (\"$ts_par\", '$stat_name', \"$stat_value\")")
					or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
			} else {
				# get previous value to calc delta
				my $delta = 0;
				my $sth1;

				$sth1 = $dbh->prepare("select value_stat_number from value_stat " . 
						"where value_stat_id = (select max(value_stat_id) " . 
						"from value_stat where value_stat_name = ?)")
					or die "ERROR: PREPARE to get previous value\n";
				$sth1->execute($stat_name);
				@delta_value = $sth1->fetchrow_array();
				$delta = $stat_value - $delta_value[0];
				$sth1->finish;

				$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
					" value_stat_number, value_stat_delta) VALUES (\"$ts_par\", '$stat_name', $stat_value, $delta)")
					or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
			} # if $stat_value
		} # while
	} else {
		$sth = $dbh->prepare("SHOW /* !50001 GLOBAL */ STATUS LIKE ?")
			or die "ERROR: SHOW GLOBAL STATUS failed\n";
		while ( <$in_stat> ) {
			chomp; my $stat_ref = $_;
			print "***> status names: *$stat_ref*\n";
			$sth->execute($stat_ref);
			while ( @mystatlist_values = $sth->fetchrow_array() ) {
				# print "$ts - Row: $mystatlist_values[0] -> $mystatlist_values[1]\n";
				$stat_name = $mystatlist_values[0];
				$stat_value = $mystatlist_values[1];
				
				if  (($stat_value =~ /\D/) || (length($stat_value) eq 0)) {
					$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
						" value_stat_string) VALUES (\"$ts_par\", '$stat_name', \"$stat_value\")")
						or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
				} else {
				# get previous value to calc delta
					my $delta;
					my $sth1;

					$sth1 = $dbh->prepare("select value_stat_number from value_stat where value_stat_id = (select max(value_stat_id) from value_stat where value_stat_name = ?)")
						or die "ERROR: PREPARE to get previous value\n";
					$sth1->execute($stat_name)
						or die "ERROR: executing selct to locate previous value for $stat_name";
					@delta_value = $sth1->fetchrow_array();
					$delta = $stat_value - $delta_value[0];
					$sth1->finish;

					$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
					  " value_stat_number, value_stat_delta) VALUES (\"$ts_par\", '$stat_name', $stat_value, $delta)")
						or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
				}
			}
		}
	}
	if ($my_all_status eq 0) {
		close $in_stat;
	}
}	# sub sub get_status_vars

# Populates all of the variable and status hashes
sub get_all_vars {
	my (@myvarlist_names, @myvarlist_values,@mystatlist_names,@mystatlist_values);
	my ($in_var);
	# We need to initiate at least one query so that our data is useable

	# Retrieve the timestamp straight from the DB
	my (@ts_out, $ts);
	@ts_out = $dbh->selectrow_array("SELECT NOW()")
		or die "ERROR: SELECT NOW() failed\n".
                       "   --> $!\n";
	$ts = $ts_out[0]; print "*** Sampled at: $ts\n";

	if ($opt{varfile} ne 0) {
		print "Using variables list '$opt{varfile}'\n";
		open($in_var,"< $opt{varfile}") or die "$opt{varfile}: $!\n";
	} else {
		print "Will retrieve all variables\n";
		$my_all_variables = 1;
	} # if opt(varfile)

	# Retrieve the list of (global) variables
	print "***> my_variables: $my_variables\n";
	if ($my_variables eq 1) {	# value is set at the begining of the main section
		print "*** Retrieving configuration variables from the DB\n";
		my ($var_name, $var_value);

		# if sepecified purge old values
		if ($my_purge ne 0) {
			$sth = $dbh->prepare("DELETE FROM value_var ".
				"WHERE DATE(value_stat_timestamp) = DATE(SUBDATE(NOW(), INTERVAL ? DAY))")
				or die "ERROR: DELETE FROM value_var on prepare\n";
			$sth = execute($my_purge)
				or die "ERROR: DELETE FROM value_var on execute\n";
		} # if my_purge

		if ($my_all_variables eq 1) {
			$sth = $dbh->prepare("SHOW /* !50001 GLOBAL */ VARIABLES")
				or die "ERROR: SHOW (GLOBAL) VARIABLES failed\n";
			$sth->execute();

			while ( @myvarlist_values = $sth->fetchrow_array() ) {
				$var_name = $myvarlist_values[0];
				$var_value = $myvarlist_values[1];
				
				if  (($var_value =~ /\D/) || (length($var_value) eq 0)) {
					$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
						" value_var_string) VALUES (\"$ts\", '$var_name', \"$var_value\")")
						or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
				} else {
					$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
						" value_var_number) VALUES (\"$ts\", '$var_name', $var_value)")
						or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
				} # if var_value
			} # while @myvarlist_values
		} else {
			$sth = $dbh->prepare("SHOW /* !50001 GLOBAL */ VARIABLES LIKE ?")
				or die "ERROR: SHOW GLOBAL VARIABLES failed\n";
			while ( <$in_var> ) {
				chomp; my $var_ref = $_;
				print "***> var names: *$var_ref*\n";
				$sth->execute($var_ref) or die "*** miserably  ***";
				while ( @myvarlist_values = $sth->fetchrow_array() ) {
					# print "$ts - Row: $myvarlist_values[0] -> $myvarlist_values[1]\n";
					$var_name = $myvarlist_values[0];
					$var_value = $myvarlist_values[1];
					
					if  (($var_value =~ /\D/) || (length($var_value) eq 0)) {
						$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
							" value_var_string) VALUES (\"$ts\", '$var_name', \"$var_value\")")
							or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
					} else {
						$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
							" value_var_number) VALUES (\"$ts\", '$var_name', $var_value)")
							or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
					} # if var_value
				} # while @myvarlist_values
			close $in_var;
			} # while
		} # if my_all_variables
	} # if my_variables

	# Retrieve the list of (global) status
	if ($my_status eq 1) {	# value is set at the begining of the main section
		print "*** Retrieving status variables from the DB\n";
		get_status_vars($ts);
	} # $my_status eq 1

} # sub get_all_vars

# ---------------------------------------------------------------------------
# BEGIN 'MAIN'
# ---------------------------------------------------------------------------
print	" * mysar v$mysarversion - Gerardo Narvaja\n".
	" * Developed at Pythian Inc. http://www.pythian.com\n".
	" * Use \"perldoc mysar.pl\" for usage informationi\n";

if ($opt{status} ne 0) { $my_status = 1; }
if ($opt{variables} ne 0) { $my_variables = 1; }
if (($my_variables eq 0) && ($my_status eq 0)) {
	print "\nERROR: You need to specify --status and/or --variables\n\n";
	usage();
	exit 1;
	}

mysql_setup();				# Gotta login first
verify_table("value_var");		# Verify the tables exist, if not create them
verify_table("value_stat");
get_all_vars();				# Get the variables and status and store them into the DB
$dbh->disconnect();			# Disconnect from the DB
print "OK - Finished successfully\n";
exit 0;
# ---------------------------------------------------------------------------
# END 'MAIN'
# ---------------------------------------------------------------------------
__END__

# ---------------------------------------------------------------------------
# pod format documentation
# ---------------------------------------------------------------------------

=head1 NAME

mysar - sar-like utility for MySQL

=head1 VERSION

Current version is v0.5-alpha
Report bugs @ http://bugs.launchpad.net/mysar/

=head1 SYNOPSIS

Minimum set of parameters

   mysar.pl --user user --pass passwd --variables|status

See B<USAGE> for optional parameters and defaults.

=head1 DESCRIPTION

This utiliy sole purpose is to take a snapshot of MySQL's running
status in a number of tables so it they can be queried and cross 
reference with other utilities and external and internal logs. For 
example: sar or MySQL's query logs.

=head1 USAGE 

 Options List:
      --host <hostname>    Connect to a remote host to perform tests
      --socket <socket>    Use a different socket for a local connection
      --port <port>        Port to use for connection (default: 3306)
      --user <username>    (required) Username to use for authentication
      --pass <password>    (required) Password to use for authentication
      --variables          Gather the global variables values
      --status             Gather the global status values
                           You need to specify at least one of the options
                           above
      --varfile            File with the list of variables to monitor
      --statfile           File with the list of status variables to 
                           monitor 
                           For variables/status specification you may 
                           use '%' as wildcard
      --database           Specifies which schema contains the tables 
                           'value_var' & 'value_stat'
      --purge              Specify the number of rows to keep in the table(s)
                           being updated.

=head2 Options details

=over 3

=item host

Specify the host being monitored, default value is B<localhost>

=item socket

Specify the socket to use to connecto to the database. The default 
value is taken from the configuration files.

=item port

Which port to use to connect to the database. Default value is B<3306>

=item user

I<MANDATORY PARAMETER>.
Which user to connect to the database. There is no default value. 
Make sure that it has been granted at least B<PROCESS>, B<CREATE> for all 
schemas and B<ALL> for the B<mysar> schema.

   GRANT USAGE, PROCESS, CREATE ON *.* TO 'mysar'@'localhost' IDENTIFIED BY 'passwd';
   GRANT ALL ON mysar.* TO 'mysar'@'localhost' 
   
User mysar should only be defined to connect from B<localhost>, unless 
you're monitoring a remote server. Replace B<mysar> and B<passwd> with 
the values of your preference.

=item pass

I<MANDATORY PARAMETER>. Password to use to connect to the database.
There is no default value.

=item variables / status

I<ONE OF THE 2 HAS TO BE PRESENT>. One of these 2 parameters needs to be 
present. B<variables> gathers the result of SHOW GLOBAL VARIABLES and 
B<status> from SHOW GLOBAL STATUS.

=item varfile

It is possible to limit the variables to gather by using a file where 
each line contains a reg expression like what you would use in 
SHOW GLOBAL VARIABLES LIKE 'reg_exp'.

=item statfile

Same as before with the expressions to use in SHOW GLOBAL STATUS LIKE 'reg_exp'.

=item database

Specify the schema that mysar will use to create its tables. Details 
on defaults are not completely defined and currently the database has 
to exist when running mysar.pl for the first time. Use:

   CREATE DATABASE dbname;

=item purge

Specify the number of days being kept in the tables that are being updated.
The default is B<30>, specifying B<0> implies no purge.
   
=back

=head1 KNOWN ISSUES

=head2 Running on other OS than Linux

I have not tested the routine with other OS like Windows and Solaris.
I currently don't see any issues on why it shouldn't run on them as
long as you have a proper Perl installation. This may vary in future
releases, especially Windows due to its file naming convention. In 
the long term Windows will be fully supported, but for now it'll be 
implemented towards the end of each release cycle (a resources issue).

=head2 Running with MySQL v4.1

Since there is no need for the trigger, it should run with MySQL v4.1.
However, since v4.1 is an I<end of life'd> release, keeping it backward
compatible with v4.1 is not a priority (another resources issue). The 
utility will remain untested for earlier releases.